"""
If you use include files for scripts add into them

    from bombard.mock_globals import *; master(<your yaml>)

That defines globals so you have valid code and code autocomplete in your IDE editor.
All strings with `bombard.examples.mock_globals` will be automatically removed before running bombard scripts.
"""
from bombard.campaign_yaml import yaml
from bombard.expand_file_name import expand_relative_file_name
import os.path # do not remove. we use it to simplify import lines in examples #NOSONAR


resp = {}


def reload(requests, repeat=None, **kwargs):
    pass


class _Supply(dict):
    pass


class _Args:
    pass


class _Ammo(dict):
    pass


ammo = _Ammo()
supply = _Supply()
args = _Args()


def master(yaml_file_name: str):
    """
    Add names from the yaml file to the unit globals.
    That makes code autocomplete work in bombard script.
    """
    try:
        campaign_book = yaml.load(open(expand_relative_file_name(yaml_file_name), 'r'))
        for name in campaign_book['ammo']:
            setattr(ammo, name, None)
        for name in campaign_book['supply']:
            setattr(supply, name, None)
    except Exception:
        pass  # ignore if the file is not completed at the moment
    args.file_name = None
    args.repeat = None
    args.verbose = None
    args.log = None
    args.ms = None
    args.threshold = None



"""
Logger with thread storage to add to log information about request.

Use main_thread/sending/receiving to switch logger modes.
"""
import threading
import logging
from bombard.terminal_colours import OFF, GRAY
from typing import Optional
from bombard.pretty_ns import time_ns, pretty_ns as plain_ns


log = logging.getLogger()
thread_data = threading.local()
pretty_ns = plain_ns  # we can redefine that to have customized formatting


def main_thread():
    """
    We are in main thread
    """
    thread_data.thread_id = 'Main'
    thread_data.request_id = ''
    thread_data.request_name = ''
    thread_data.dir = ''
    thread_data.start = None
    thread_data.colour = OFF


def sending(thread_id, request_id, request_name):
    """
    Start sending request
    """
    thread_data.thread_id = thread_id
    thread_data.request_id = request_id
    thread_data.request_name = request_name
    thread_data.dir = '>' * 3
    thread_data.colour = GRAY
    thread_data.start = time_ns()


def receiving():
    """
    Got response to request
    """
    thread_data.dir = '<' * 3
    thread_data.colour = OFF


class RequestFormatter(logging.Formatter):
    def format(self, record):
        record.threadid = thread_data.thread_id
        record.requestid = str(thread_data.request_id).rjust(4)
        record.colour = thread_data.colour
        record.requestname = thread_data.request_name
        record.dir = thread_data.dir
        record.elapsed = pretty_ns(time_ns() - thread_data.start) if thread_data.start is not None else ''
        return super().format(record)


def setup_logging(level: int, log_file_name: Optional[str] = None):
    main_thread()
    handler = logging.StreamHandler()
    formatter = RequestFormatter(
            fmt=f'%(colour)s%(asctime)s %(requestid)s %(elapsed)s (thread %(threadid)s) '
                f'%(requestname)s %(dir)s %(message)s{OFF}',
            datefmt='%d %b %H:%M:%S'
    )
    handler.setFormatter(formatter)
    handler.setLevel(level)
    log.handlers = []
    log.addHandler(handler)
    log.setLevel(logging.DEBUG)

    if log_file_name is not None:
        file_handler = logging.FileHandler(log_file_name)
        file_handler.setFormatter(formatter)
        file_handler.setLevel(logging.DEBUG)
        log.addHandler(file_handler)



"""
Fake args that we can pass to bombard to test it.
This is AttrDict so you can use something like that
>>> args = FakeArgs(init=True)
>>> args.init
True
"""
from bombard.attr_dict import AttrDict
from bombard.args import CAMPAIGN_FILE_NAME, THRESHOLD, TIMEOUT, THREADS_NUM


class FakeArgs(AttrDict):
    threads = THREADS_NUM
    timeout = TIMEOUT
    ms = False
    threshold = THRESHOLD
    quiet = False
    dry = False
    verbose = False
    quiet = False
    log = None
    example = None
    init = False
    examples = False
    file_name = CAMPAIGN_FILE_NAME
    version = False
    supply={}
    repeat=1



import unittest
from bombard import bombardier


TEST_REQUEST = {
    'url': 'https://localhost/users',
    'method': 'GET',
    'headers': {'x-x': 'json'}
}

TEST_REQUEST_TEMPLATE = {
    '1': {
        '2': '{a}'
    }
}
TEST_SUPPLY = {'a': '4'}


class TestHttpRequests(unittest.TestCase):
    def testApplySupply(self):
        self.assertEqual(bombardier.apply_supply_dict(TEST_REQUEST_TEMPLATE, TEST_SUPPLY)['1']['2'], '4')


if __name__ == '__main__':
    unittest.main()


"""
- time_ns for Python before 3.7
- Elapsed context manager
- pretty_ns to represent time elapsed in human lovable form

Usage:
>>> start = time_ns()
>>> import time
>>> time.sleep(0.0000001)
>>> pretty_ns((start + 100) - start)
'0.1 mks'

>>> with Timer() as timer:
...     import time
...     time.sleep(0.0000001)
...     timer.pretty.endswith('s')
...     timer.ns > 0
True
True

"""

time_ns = None  # for Python3.7+ this is function from system library time


# for earlier Python versions this is emulation of the Python3.7 time_ns


def pretty_ns(elapsed_ns: int, fixed_units=None):
    dividers = {
        'us': 1,
        'mks': 1000,
        'ms': 1000,
        'sec': 1000,
        'minutes': 60,
        'hours': 60,
        'days': 24
    }
    result = elapsed_ns
    for unit, divider in dividers.items():
        result /= divider
        if result < 100 or unit.lower() == fixed_units:
            return f'{result:.1f} {unit}'
        else:
            result = round(result)


def emul_time_ns():
    return int(perf_counter() * 10 ** 9)


class Timer:
    def __init__(self):
        pass

    def __enter__(self):
        self.start = time_ns()
        return self

    def __exit__(self, *args):
        pass

    @property
    def ns(self):
        return time_ns() - self.start

    @property
    def pretty(self):
        return pretty_ns(self.ns)


import time

try:
    time_ns = time.time_ns
except AttributeError:
    from time import perf_counter

    time_ns = emul_time_ns

if __name__ == "__main__":
    import doctest

    doctest.testmod()



from tests.stdout_capture import CaptureOutput
import unittest
from bombard.report import Reporter


class TestInit(unittest.TestCase):
    def setUp(self):
        self.reporter = Reporter(
            )

    def testZeroTime(self):
        """
        zero time => infinity numbers
        """
        report = self.reporter.report()
        self.assertIn(  
                '\u221E',  # utf8 infinity char ∞
                report
                )



from tests.stdout_capture import CaptureOutput
import unittest
from bombard.main import campaign
from tests.fake_args import FakeArgs
from bombard.args import INIT_EXAMPLE, CAMPAIGN_FILE_NAME
import os.path


def clean_campaign_file():
    """ Removes default campaign from folder root """
    if os.path.isfile(CAMPAIGN_FILE_NAME):
        os.remove(CAMPAIGN_FILE_NAME)


class TestInit(unittest.TestCase):
    def setUp(self):
        clean_campaign_file()

    def tearDown(self):
        clean_campaign_file()

    def testInitDefault(self):
        with CaptureOutput() as captured:
            campaign(FakeArgs(init=True))

        self.maxDiff = 1024
        with open(f'bombard/examples/{INIT_EXAMPLE}') as ex, open(f'{CAMPAIGN_FILE_NAME}', 'r') as init:
            self.assertEqual(
                ex.read(),
                init.read()
                )

    def testInitSimpleton(self):
        with CaptureOutput() as captured:
            campaign(FakeArgs(example='simpleton', init=True))

        self.maxDiff = 1024
        with open(f'bombard/examples/simpleton.yaml') as desc, open(f'{CAMPAIGN_FILE_NAME}', 'r') as init:
            self.assertIn(
                desc.read(),
                init.read()
                )


import unittest
from unittest.mock import MagicMock
from bombard import bombardier
from bombard import http_request
from bombard.main import setup_logging
import logging
from tests.fake_args import FakeArgs


TEST_AMMO = {
    'request': {
        'url': 'https://localhost/users',
        'method': 'GET',
        'headers': {'x-x': 'json'}
    },
    'id': 1,
    'name': 'request',
    'supply': {},
}


class TestHttpRequests(unittest.TestCase):
    def setUp(self):
        http_request.http.client.HTTPSConnection.request = MagicMock()
        http_request.http.client.HTTPSConnection.getresponse = MagicMock()
        setup_logging(logging.DEBUG)

    def testRequest(self):
        bombardier.Bombardier(args=FakeArgs()).worker(1, TEST_AMMO)
        http_request.http.client.HTTPSConnection.request.assert_called_once_with(
            'GET', '/users', body=None, headers={'x-x': 'json'}
        )
        http_request.http.client.HTTPSConnection.getresponse.assert_called_once()


if __name__ == '__main__':
    unittest.main()



"""
Multi-threaded jobs processor abstraction.

Override method worker in descendant to do a job.
Add jobs with `put`.
Start processing with `start`.
"""
from threading import Thread
from queue import Queue
from copy import deepcopy
from abc import abstractmethod


class WeaverMill:
    def __init__(self, threads_num: int=10):
        """
        :param threads_num: How many threads we will use.
        """
        self.threads_num = threads_num
        self.threads = []
        self.queue = Queue()
        self.job_count = 0
        for thread_id in range(threads_num):
            t = Thread(target=self.thread_worker, args=[thread_id])
            t.daemon = True
            t.start()
            self.threads.append(t)

    def thread_worker(self, thread_id):
        """
        Get job from queue and pass it to abstract worker
        that should be implemented in descendant.

        Even if worker throw exception we mark the job we gave him as done.

        Job == None is a signal to stop work.
        """
        while True:
            job = deepcopy(self.queue.get())
            if job is None:
                break
            try:
                self.worker(thread_id, job)
            finally:
                self.queue.task_done()

    @abstractmethod
    def worker(self, thread_id, job):
        """
        Implement your job processor, runs in thread.

        :param thread_id: just sequential number of the thread we work into
        :param job: job from queue
        """
        pass

    def put(self, job):
        """
        Add job to queue.
        To start processing use `process`.
        """
        self.queue.put(job)

    def process(self):
        """ Starts all threads and lock until queue is empty """
        self.queue.join()

    def stop(self):
        """ Stops all threads - send stop signal to queue and lock until they stop """
        for _ in range(len(self.threads)):
            self.queue.put(None)
        for thread in self.threads:
            thread.join()



"""
Dict with access to keys as attributes.

>>> d = AttrDict({'a': 'b', 'c': 'g'})
>>> d.c
'g'

Instance can be used as callable to set value to the dict
>>> d = AttrDict({})
>>> d(k='t')
>>> d.k
't'

(!) All values that you set will be repeated to `master` that you used to init the object.
>>> master = {'j': 'l'}
>>> d = AttrDict(master)
>>> d(h='q')
>>> master['h']
'q'
>>> d['w'] = 'e'
>>> master['w']
'e'

And it works without master
>>> d = AttrDict()
>>> d(f=7)
>>> d.f
7

"""


class AttrDict(dict):
    """
    You can access all dict values as attributes.
    All changes immediately repeated in master_dict.
    """
    def __call__(self, **kwargs):
        for name, val in kwargs.items():
            self[name] = val

    def __init__(self, master_dict: dict=None, **kwargs):
        if master_dict is None:
            super().__init__(**kwargs)
        else:
            super().__init__(master_dict, **kwargs)
        self.__dict__ = self
        self.master_dict = master_dict

    def __setitem__(self, name, val):
        if self.master_dict is not None:
            self.master_dict[name] = val
        super().__setitem__(name, val)  # __dict__ pointed to self so all dict items became attributes


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import json


FAKE_RESP = {
    ('GET', '/posts'): [
        {'id': 1}, {'id': 2}, {'id': 3}
    ],
    ('GET', '/posts/1'): '',
    ('GET', '/posts/2'): '',
    ('GET', '/posts/3'): '',
}


class FakeResp:
    status = 200
    body = None

    def __init__(self, resp_body):
        self.body = resp_body

    def read(self):
        return self.body


class FakeJSONPlaceholder:
    def request(self, method, path, body=None, headers=None):
        self.method = method
        self.path = path

    def response(self):
        return FakeResp(json.dumps(
            FAKE_RESP[(self.method, self.path)]
        ))



"""
Capture stdout and stderr.

Since Python 3.4 we have redirect_std* in contextlib.
But I still prefer my context manager as simpler to use.

Usage:
>>> with CaptureOutput() as captured:
...     print('3', end='')
>>> captured.output
'3'

>>> with CaptureOutput(capture=False) as captured:
...     captured.output is None
True
"""
import sys
from io import StringIO


class CaptureOutput:
    def __init__(self, capture=True):
        """ To see output you can set capture to False"""
        self.capture = capture

    def __enter__(self):
        self.old_out, self.old_err = sys.stdout, sys.stderr
        if self.capture:
            sys.stdout = self.out = StringIO()
            sys.stderr = self.err = StringIO()
        return self

    @property
    def stdout(self):
        if self.capture and self.out.getvalue():
            return self.out.getvalue()
        else:
            return None

    @property
    def stderr(self):
        if self.capture and self.err.getvalue():
            return self.err.getvalue()
        else:
            return None

    @property
    def output(self):
        """ stdout and strerr separated by new line """
        if self.stdout is not None and self.stderr is not None:
            return '\n'.join([self.stdout, self.stderr])
        elif self.stdout is not None:
            return self.stdout
        else:
            return self.stderr

    def __exit__(self, *args):
        if self.capture:
            sys.stdout, sys.stderr = self.old_out, self.old_err


if __name__ == "__main__":
    import doctest
    doctest.testmod()


"""
Bombard campaign loader.

Extends yaml loader with loading external files `!include file.ext`.
Excludes lines that import mock_globals.
"""
import os.path
import yaml as original_yaml
import io


SIGNATURE = 'bombard.mock_globals'


class Yaml:
    @staticmethod
    def load(*args, **kwargs):
        """
        Mimics yaml interface for seamless injection
        """
        return original_yaml.load(*args, **kwargs, Loader=IncludesLoader)

    @staticmethod
    def full_load(*args, **kwargs):
        """
        Mimics yaml interface for seamless injection
        """
        return original_yaml.load(*args, **kwargs, Loader=IncludesLoader)


yaml = Yaml()


class IncludesLoader(original_yaml.SafeLoader):
    def __init__(self, stream):
        self._root = os.path.split(stream.name)[0]
        super(IncludesLoader, self).__init__(stream)

    @staticmethod
    def wrap_in_yaml_document(msg: str) ->str:
        """
        Converts multi-line msg to yaml document that we can insert into yaml
        """
        result = []
        for line in msg.split('\n'):
            if SIGNATURE not in line:  # skip line that mocks globals
                result.append(' '*4 + line)
        return '|\n' + '\n'.join(result)

    def include(self, node):
        filename = os.path.join(self._root, self.construct_scalar(node))
        with open(filename, 'r') as f:
            wrapped = io.StringIO(self.wrap_in_yaml_document(f.read()))
            wrapped.name = f.name  # to please owe own __init__
            return original_yaml.load(wrapped, IncludesLoader)


IncludesLoader.add_constructor('!include', IncludesLoader.include)




"""
doctests from all modules of packages listed in PACKAGES
will be added to unit tests suite.

So you will have one nice test report and one test result/exit code in CI.

You should import below all your packages with doctests and list them in PACKAGES.
"""
import doctest
from os.path import dirname, basename, isfile
from glob import glob
from importlib import import_module

import tests
import bombard
PACKAGES = [bombard, tests]

PY_EXT = '.py'


def load_tests(loader, tests, ignore):
    """
    Unittest hook to add tests to auto-discovered ones
    https://docs.python.org/3/library/unittest.html#load-tests-protocol
    """
    for package in PACKAGES:
        for module_file in glob(dirname(package.__file__) + f'/*{PY_EXT}'):
            if isfile(module_file) and not module_file.endswith('__init__.py'):
                tests.addTest(doctest.DocTestSuite(
                    import_module(f'{package.__name__}.{basename(module_file)[:-len(PY_EXT)]}')
                ))
    return tests



import http.client
import ssl
from urllib.parse import urlparse
from typing import Optional


EXCEPTION_STATUS = '!!!'


def http_request(url: str, method: str='GET', headers: Optional[dict]=None,
                 body: Optional[str]=None, timeout: Optional[int]=None) -> (int, dict):
    """
    Make HTTP request.

    Returns tuple:
        <HTTP response status>, <response body>
    """
    try:
        url = urlparse(url)
        kwargs = {'timeout': timeout} if timeout is not None else {}
        if url.scheme.lower() == 'https':
            conn = http.client.HTTPSConnection(
                url.netloc,
                context=ssl._create_unverified_context(),
                **kwargs,
            )
        else:
            conn = http.client.HTTPConnection(
                url.netloc,
                **kwargs,
            )
        conn.request(
            method,
            url.path,
            body=body,
            headers=headers if headers is not None else {}
        )
        resp = conn.getresponse()
        resp_body = resp.read()
    except Exception as e:
        return EXCEPTION_STATUS, str(e)
    return resp.status, resp_body




import unittest
from bombard.request_logging import setup_logging
import logging
from tests.stdout_capture import CaptureOutput
from tests.fake_args import FakeArgs
from bombard.main import start_campaign


EMPTY_BOOK = {}

class TestCampaignCheck(unittest.TestCase):
    def setUp(self):
        setup_logging(logging.DEBUG)

    def testSimplest(self):
        with CaptureOutput(capture=True) as captured:
            start_campaign(FakeArgs(example='simplest'), EMPTY_BOOK)
        self.assertIn(
            'You should have at least one of "prepare" and "ammo"',
            captured.output
        )




import unittest
from bombard import http_request
from bombard.request_logging import setup_logging
import logging
from tests.stdout_capture import CaptureOutput
from tests.fake_args import FakeArgs
from bombard.main import campaign
from tests.fake_jsonplaceholder import FakeJSONPlaceholder


class TestExamples(unittest.TestCase):
    def setUp(self):
        self.fake_server = FakeJSONPlaceholder()
        http_request.http.client.HTTPSConnection.request = self.fake_server.request
        http_request.http.client.HTTPSConnection.getresponse = self.fake_server.response
        setup_logging(logging.DEBUG)

    def testSimplest(self):
        with CaptureOutput(capture=True) as captured:
            campaign(FakeArgs(example='simplest'))
        self.assertIn(
            'GET jsonplaceholder.typicode.com/posts',
            captured.output
        )
        self.assertIn(
            'GET jsonplaceholder.typicode.com/posts/1',
            captured.output
        )



def pretty_sz(size):
    dividers = {
        'bytes': 1,
        'kb': 1024,
        'mb': 1024,
        'gb': 1024,
        'pb': 1024,
    }
    result = size
    for unit, divider in dividers.items():
        result /= divider
        if result < 100:
            return f'{result:.1f} {unit}'
        else:
            result = round(result)



"""
Parse bombard command line args.
"""
import argparse

from bombard.show_descr import markdown_for_terminal
from bombard.terminal_colours import BROWN, OFF

# from pkg_resources import resource_string


EXAMPLES_PREFIX = 'bombard://'  # replaced with path to package folder
DIR_DESC_FILE_NAME = 'README.md'  # if directory as campaign file then show content of this file from the directory
THREADS_NUM = 10
CAMPAIGN_FILE_NAME = 'bombard.yaml'
INIT_EXAMPLE = 'easy.yaml'
REPEAT = 10
THRESHOLD = 1000
TIMEOUT = 15


def get_args():
    parser = argparse.ArgumentParser(
        description=markdown_for_terminal(f'''bombard: utility to bombard with HTTP-requests.

{BROWN}[GitHub](https://github.com/masterandrey/bombard){OFF}'''),
        epilog=markdown_for_terminal('''To show available examples use `bombard --examples`''')
    )
    parser.add_argument(
        dest='file_name', type=str, nargs='?',
        default=CAMPAIGN_FILE_NAME,
        help=f'''file name with bombing campaign plan (default "#{CAMPAIGN_FILE_NAME}").
To use bombard examples prefix filename with "@".'''
    )
    parser.add_argument(
        '--parallel', '-p', dest='threads', type=int,
        default=THREADS_NUM,
        help=f'number of simultaneous requests (default {THREADS_NUM})'
    )
    parser.add_argument(
        '--supply', '-s', dest='supply', type=str, nargs='*',
        help='supply as separate pairs "-c name=val" or many pairs at once "-c name1=val1,name2=val2,.."'
    )
    parser.add_argument(
        '--repeat', '-r', dest='repeat', type=int, default=REPEAT,
        help=f'how many times to repeat (by default {REPEAT})'
    )
    parser.add_argument(
        '--verbose', '-v', dest='verbose', default=False, action='store_true',
        help=f'verbose output (by default False)'
    )
    parser.add_argument(
        '--version', dest='version', default=False, action='store_true',
        help=f'bombard version'
    )
    parser.add_argument(
        '--log', '-l', dest='log', type=str, default=None,
        help=f'log file name'
    )
    parser.add_argument(
        '--ms', '-m', dest='ms', default=False, action='store_true',
        help=f'Show all times in ms (by default use intellectual format)'
    )
    parser.add_argument(
        '--threshold', '-t', dest='threshold', type=int, default=THRESHOLD,
        help=f'threshold in ms. all times greater than that will be shown in red (default {THRESHOLD})'
    )
    parser.add_argument(
        '--timeout', dest='timeout', type=int, default=TIMEOUT,
        help=f'http timeout in seconds (default {TIMEOUT})'
    )
    parser.add_argument(
        '--quiet', '-q', dest='quiet', default=False, action='store_true',
        help=f'suppress printing request/response to improve performance'
    )
    parser.add_argument(
        '--example', '-e', dest='example', type=str, default=None,
        help=f'''get bombard campaign from internal bombard example with the name. 
to list all available examples use `--examples`.'''
    )
    parser.add_argument(
        '--examples', '-x', dest='examples', default=False, action='store_true',
        help=f'''show all available examples description.'''
    )
    parser.add_argument(
        '--dry', '-d', dest='dry', default=False, action='store_true',
        help=f'run without actual HTTP requests. if there is "dry" parameter in an ammo use it as fake request result.'
    )
    parser.add_argument(
        '--init', '-i', dest='init', default=False, action='store_true',
        help=f'''copy `{INIT_EXAMPLE}` example to current folder with the name {CAMPAIGN_FILE_NAME} 
so it will be used as default with `bombard` command. If you want to copy different example add option
`--example <the name you want>`. For full list of examples use `--examples` option.
'''
    )

    args = parser.parse_args()
    args.parser = parser

    return args





import json
import logging
from typing import Mapping
from urllib.parse import urlparse

from bombard import request_logging
from bombard.attr_dict import AttrDict
from bombard.http_request import http_request, EXCEPTION_STATUS
from bombard.pretty_ns import time_ns
from bombard.pretty_sz import pretty_sz
from bombard.report import Reporter
from bombard.show_descr import markdown_for_terminal
from bombard.terminal_colours import red, dark_red, green
from bombard.weaver_mill import WeaverMill

log = logging.getLogger()

DEFAULT_OK = [200]
DEFAULT_OVERLOAD = [502, 504]
PREPARE = 'prepare'
AMMO = 'ammo'


def apply_supply(s: str, supply: dict) -> str:
    # todo: add args
    if not isinstance(s, str):
        return s
    try:
        return eval('f"""' + s + '"""', supply)
    except Exception as e:
        log.error(f'Cannot eval "{s}":\n{e}', exc_info=True)
    return s


def apply_supply_dict(request: dict, supply: dict) -> dict:
    """
    Use supply to substitute all {name} in request strings.
    """
    for name in request:
        if isinstance(request[name], dict):
            request[name] = apply_supply_dict(request[name], supply)
        elif isinstance(request[name], str):
            request[name] = apply_supply(request[name], supply)
    return request


class Bombardier(WeaverMill):
    """
    Use horde of threads to make HTTP-requests
    """

    def __init__(self, supply: dict = None, args=None, campaign_book: dict = None, ok_statuses=None,
                 overload_statuses=None):
        self.supply = supply if supply is not None else {}
        self.supply['args'] = args
        self.args = args
        self.campaign = campaign_book
        self.ok = ok_statuses if ok_statuses is not None else DEFAULT_OK
        self.overload = overload_statuses if overload_statuses is not None else DEFAULT_OVERLOAD
        self.request_fired = False  # from any request reload() was called
        self.resp_count = 0

        self.show_request = {
            1: 'Sent 1st request..'
        }
        self.show_response = {
            1: 'Got 1st response..'
        }
        self.reporter = Reporter(
            time_units=('ms' if args.ms else None),
            time_threshold_ms=int(args.threshold),
            success_statuses=self.ok
        )
        request_logging.pretty_ns = self.reporter.pretty_ns

        super().__init__(args.threads)

    def status_coloured(self, status: int) -> str:
        if status in self.ok:
            return green(str(status))
        elif status in self.overload:
            return dark_red(str(status))
        else:
            return red(str(status))

    @staticmethod
    def get_headers(request: dict, body_is_json: bool) -> dict:
        """
        Treat special value 'json' as Content-Type: application/json
        """
        predefined = {
            'json': {'Content-Type': 'application/json'},
        }
        if 'headers' in request and isinstance(request['headers'], str):
            # if headers in request description just a string we know this should be some predefined code
            for known in predefined:
                if request['headers'].lower() == known:
                    return predefined[known]
        result = {}
        for name, val in request.get('headers', {}).items():
            for known in predefined:
                if name.lower() == known:
                    result.update(predefined[known])
                    break
            else:
                result.update({name: val})
        if body_is_json and 'Content-Type' not in result:
            result.update(predefined['json'])
        return result

    def process_resp(self, ammo: dict, status: int, resp: str, elapsed: int, size: int):
        request = ammo['request']
        self.reporter.log(status, elapsed, request.get('name'), size)
        if status in self.ok:
            log.debug(f'{status} reply\n{resp}')
            if 'extract' in request:
                try:
                    data = json.loads(resp)
                    if not hasattr(request['extract'], 'items'):
                        request['extract'] = {request['extract']: request['extract']}
                    for name, extractor in request['extract'].items():
                        if not extractor:
                            extractor = name
                        if '[' in extractor:
                            self.supply[name] = eval('data' + extractor)
                        else:
                            self.supply[name] = data[extractor]
                except Exception as e:
                    log.error(f'Cannot extract {request["extract"]} from {resp}:\n{e}', exc_info=True)
            if 'script' in request:
                try:
                    # Supply immediately repeats all changes in the self.supply so if the script spawns new
                    # requests they already get new values
                    supply = AttrDict(self.supply, **ammo['supply'])
                    context = {
                        'reload': self.reload,
                        'resp': json.loads(resp),
                        'args': self.args,
                        'supply': supply,
                        'ammo': AttrDict(self.campaign[AMMO])
                    }
                    if 'compiled' not in request:
                        request['compiled'] = compile(request['script'], 'script', 'exec')
                    exec(request['compiled'], context)
                except Exception as e:
                    log.error(f'Script fail\n{e}\n\n{request["script"]}\n\n{supply}\n', exc_info=True)

    @staticmethod
    def beautify_url(url, method, body):
        urlparts = urlparse(url)
        path = urlparts.path if len(urlparts.path) < 15 else '...' + urlparts.path[-15:]
        query = '?' + urlparts.query if urlparts.query else ''
        if urlparts.fragment:
            query += '#' + urlparts.fragment
        query = query if len(query) < 15 else '?...' + query[-15:]
        return f"""{method} {urlparts.netloc}{path}{query}"""

    def worker(self, thread_id, ammo):
        """
        Thread callable.
        Strike ammo from queue.
        """
        try:
            # setup logging ASAP and as safe as possible
            if isinstance(ammo, Mapping):
                request = ammo.get('request', {})
                ammo_id = ammo.get('id', '')
                ammo_name = request.get('name', '')
            else:
                request = {}
                ammo_id = None
                ammo_name = None
            request_logging.sending(thread_id, ammo_id, ammo_name)
            pretty_url = ''  # we use it in `except`
            try:
                ammo = apply_supply_dict(ammo, dict(self.supply, **ammo['supply']))

                url = request.get('url', '')
                method = request['method'] if 'method' in request else 'GET'
                body = json.dumps(request['body']) if 'body' in request else None
                headers = self.get_headers(request, body is not None)
                pretty_url = self.beautify_url(url, method, body)

                log.debug(f'Bomb to drop:\n{pretty_url}' + ('\n{body}' if body is not None else ''))
                if self.args.quiet:
                    if ammo_id in self.show_request:
                        print(f'{self.show_request[ammo_id].format(id=ammo_id):>15}\r', end='')
                log.info(pretty_url)

                start_ns = time_ns()
                if self.args.dry:
                    status, resp = self.ok[0], json.dumps(request.get('dry'))
                else:
                    status, resp = http_request(url, method, headers, body, self.args.timeout)

                request_logging.receiving()

                self.process_resp(ammo, status, resp, time_ns() - start_ns, len(resp))

                self.resp_count += 1
                if self.args.quiet:
                    if self.resp_count in self.show_response:
                        print(f'{self.show_response[self.resp_count].format(id=self.resp_count):>15}\r', end='')
                log.info(self.status_coloured(status) + f' ({pretty_sz(len(resp))}) ' + pretty_url
                         + ' ' + (red(resp) if status == EXCEPTION_STATUS else '')
                         )
            except Exception as e:
                log.info(pretty_url + ' ' + red(str(e)), exc_info=True)
        finally:
            request_logging.main_thread()

    def reload(self, requests, repeat=None, prepare=False, **kwargs):
        """
        Add request(s) to the bombardier queue `repeat`-times (args.repeat if None).
        If `repeat` field exists in the request additionally repeats as defined by it.

        Requests can be one request or list of requests.
        If supply specified it'll be used in addition to self.supply.

        Arg `prepare` indicate call from main, not from request script.
        So we know if any scripts call reload (self.request_fired)
        """
        if not prepare:
            self.request_fired = True
        if not isinstance(requests, list):
            requests = [requests]
        if repeat is None:
            repeat = self.args.repeat
        for request in requests:
            try:
                _repeat = int(apply_supply(request['repeat'], self.supply))
            except (ValueError, KeyError):
                _repeat = repeat
            for _ in range(_repeat):
                self.job_count += 1
                if self.job_count % self.args.threads == 0 \
                        or self.job_count < self.args.threads and self.job_count % 10 == 0:
                    # show each 10th response before queue is full and then each time it's full
                    self.show_response[self.job_count] = f'Got {self.job_count} responses...'
                self.put({
                    'id': self.job_count,
                    'request': request,
                    'supply': kwargs
                })

    def report(self):
        log.warning(
            '\n'
            + '=' * 100
            + '\n'
            + markdown_for_terminal(self.reporter.report())
            + '=' * 100
            + '\n'
        )



"""
Colourize text in terminal
source https://en.wikipedia.org/wiki/ANSI_escape_code#Colors

You can use it function style
>>> green('Hello!')
'\\x1b[1;32mHello!\\x1b[0m'

Or include style
>>> f'{YELLOW}Hello{OFF}, {RED}world{OFF}!'
'\\x1b[1;33mHello\\x1b[0m, \\x1b[1;31mworld\\x1b[0m!'

Under the hood this is colorama.
But I keep my wrapper in this module as legacy.
"""

from colorama.ansi import code_to_chars, AnsiFore, AnsiStyle


GREEN = code_to_chars(f'{AnsiStyle.BRIGHT};{AnsiFore.GREEN}')
RED = code_to_chars(f'{AnsiStyle.BRIGHT};{AnsiFore.RED}')
DARK_RED = code_to_chars(f'{AnsiStyle.DIM};{AnsiFore.RED}')
GRAY = code_to_chars(f'{AnsiStyle.BRIGHT};{AnsiFore.BLACK}')
BROWN = code_to_chars(f'{AnsiStyle.DIM};{AnsiFore.YELLOW}')
YELLOW = code_to_chars(f'{AnsiStyle.BRIGHT};{AnsiFore.YELLOW}')

OFF = code_to_chars(AnsiStyle.RESET_ALL)


def paint_it(msg: str, colour: str) -> str:
    return f'{colour}{msg}{OFF}'


def green(s: str) -> str:
    return paint_it(s, GREEN)


def red(s: str) -> str:
    return paint_it(s, RED)


def dark_red(s: str) -> str:
    return paint_it(s, DARK_RED)


def gray(s: str) -> str:
    return paint_it(s, GRAY)


def brown(s: str) -> str:
    return paint_it(s, BROWN)


def yellow(s: str) -> str:
    return paint_it(s, YELLOW)


if __name__ == "__main__":
    import doctest
    fail, total = doctest.testmod(report=True)
    if not fail:
        print(f'... {total} test(s) passed')



"""
Bombard's main
"""
from bombard.campaign_yaml import yaml
from bombard.bombardier import Bombardier, AMMO, PREPARE
import logging
from bombard.args import get_args, CAMPAIGN_FILE_NAME, INIT_EXAMPLE, EXAMPLES_PREFIX
from typing import Optional
from bombard.request_logging import setup_logging, log
import os.path
from shutil import copyfile
from bombard.expand_file_name import get_campaign_file_name, show_folder, expand_relative_file_name
from bombard.terminal_colours import red, RED, OFF
import bombard
import colorama


colorama.init()  # On Windows will filter ANSI escape sequences out of any text sent to 
# stdout or stderr, and replace them with equivalent Win32 calls.


def guess_type(value: str):
    """
    Converts value in int or float if possible
    """
    try:
        return int(value)
    except ValueError:
        pass
    try:
        return float(value)
    except ValueError:
        pass
    return value


def get_supply_from_cli(supply: Optional[list]):
    result = {}
    if supply:
        for item in supply:
            for var in item.split(','):
                name, val = var.split('=')
                result.update({name: guess_type(val)})
    return result


def load_book_supply(cli_supply, book_supply):
    """
    Updates CLI supply with supply from campaign book.
    But do not overwrite CLI supply with book supply - values from CLI have bigger priority.
    """
    for var, val in book_supply.items():
        if var not in cli_supply:
            cli_supply[var] = val.format(**cli_supply) if isinstance(val, str) else val


def add_names_to_requests(campaign_book):
    """
    Duplicate names inside requests so worker will see it and use in stat report
    """
    for request_set in [PREPARE, AMMO]:
        if request_set in campaign_book:
            for name, request in campaign_book[request_set].items():
                request['name'] = name


def init(args):
    """
    Copies the example to current folder as bombard.yaml
    """
    if args.example is None:
        src = expand_relative_file_name(f'{EXAMPLES_PREFIX}{INIT_EXAMPLE}')
    else:
        src = get_campaign_file_name(args)  # actually it will be from ==example
    if os.path.isfile(CAMPAIGN_FILE_NAME):
        print(f'Cannot init from {src}:\n{RED}File {CAMPAIGN_FILE_NAME} already exists.{OFF}')
        return
    copyfile(src, CAMPAIGN_FILE_NAME)
    # todo copy external python scripts if it is included into the example (create yaml CopyLoader)


def start_campaign(args, campaign_book):
    log.debug(f'Starting bombard campaign with args\n' + ' ' * 4 + f'{args.__dict__}')
    log.debug(f'Loaded bombard campaign from "{args.file_name}": {len(campaign_book.get("ammo", {}))} ammo.')
    if PREPARE not in campaign_book and AMMO not in campaign_book:
        print(f'You should have at least one of "{PREPARE}" and "{AMMO}" '
              f'section in your campaign file {args.file_name}')
        return

    supply = get_supply_from_cli(args.supply)
    load_book_supply(supply, campaign_book.get('supply', {}))
    log.debug(f'Supply: {supply}')

    add_names_to_requests(campaign_book)

    bombardier = Bombardier(supply, args, campaign_book)
    if PREPARE in campaign_book:
        for ammo in campaign_book[PREPARE].values():
            bombardier.reload(ammo, repeat=1, prepare=True)
        bombardier.process()
    if not bombardier.request_fired and AMMO in campaign_book:
        for ammo in campaign_book[AMMO].values():
            bombardier.reload(ammo, repeat=args.repeat)
        bombardier.process()
    bombardier.stop()
    bombardier.report()


def campaign(args):
    if args.version:
        print(bombard.__name__, bombard.version())
        with open(os.path.join(os.path.dirname(bombard.__file__), 'LICENSE.txt'), 'r') as license:
            print(license.readline())
        return

    if args.quiet:
        level = logging.WARNING
    elif args.verbose:
        level = logging.DEBUG
    else:
        level = logging.INFO

    setup_logging(level=level, log_file_name=args.log)

    if args.init:
        init(args)
        return

    campaign_file_name = get_campaign_file_name(args)
    if os.path.isdir(campaign_file_name):
        show_folder(campaign_file_name)
    elif not os.path.isfile(campaign_file_name) and not args.init:
        print(red(f'\nCannot find campaign file "{args.file_name}"\n'))
    else:
        start_campaign(args, yaml.load(open(campaign_file_name, 'r')))


def main():
    campaign(get_args())


if __name__ == '__main__':
    main()



from tests.stdout_capture import CaptureOutput
import unittest
from bombard.main import campaign
from tests.fake_args import FakeArgs
from bombard.args import INIT_EXAMPLE, DIR_DESC_FILE_NAME, CAMPAIGN_FILE_NAME
from bombard.show_descr import markdown_for_terminal


class TestShowFolder(unittest.TestCase):
    def testShowFolder(self):
        with CaptureOutput() as captured:
            campaign(FakeArgs(examples=True))

        self.maxDiff = 1024
        with open(f'bombard/examples/{DIR_DESC_FILE_NAME}') as desc:
            self.assertIn(  # do not check 1st line of output with the folder name
                markdown_for_terminal(desc.read()),
                captured.output
                )



from pygments import highlight
from pygments.lexers.markup import MarkdownLexer
from pygments.formatters.terminal import TerminalFormatter


def markdown_for_terminal(descr: str) -> str:
    return highlight(descr, MarkdownLexer(), TerminalFormatter())



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()    

import sys; sys.path.insert(0, '../'); import bombard

setuptools.setup(
    name='bombard',
    version=bombard.version(),
    # scripts=['bin/bombard'],
    entry_points={
        'console_scripts': [
            'bombard=bombard.main:main',
        ],
    },
    author="Andrey Sorokin",
    author_email="filbert@yandex.ru",
    description="Bombards target server with simultaneous requests",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://bombard.readthedocs.io/en/latest/",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=requirements,
    keywords='http load test parallel',
    classifiers=[
     "Programming Language :: Python :: 3",
     "License :: OSI Approved :: MIT License",
     "Operating System :: OS Independent",
    ],
 )


"""
Bombard reporter.

Use:
* `log` to add each request result.
* `report` to generate report.
"""
from bombard.pretty_ns import time_ns, pretty_ns
from array import array
import statistics
from bombard.pretty_sz import pretty_sz
from bombard.terminal_colours import red
from copy import deepcopy


ARRAY_UINT64 = 'Q'
ARRAY_UINT32 = 'L'
SUCCESS_GROUP = 'success'
FAIL_GROUP = 'fail'
GROUPS = [SUCCESS_GROUP, FAIL_GROUP]
TIME = 'time'
SIZE = 'size'
TIME_DENOMINATOR = 1  # if 1 we store ns.


class Reporter:
    """
    Report bombard's result

    Access to stat data only by log(), filter(), reduce()
    """
    def __init__(
            self,
            time_units: str='ms',
            time_threshold_ms: int=1,
            success_statuses: dict=[200]
    ):
        """
        :param time_units: time representation fixed in the units (see names in bombard.pretty_ns)
        :param time_threshold_ms: show times bigger than that in red
        :param success_statuses: dict of statuses treated as success
        """
        self.DIMENSIONS = {
            TIME: {'type': ARRAY_UINT64, 'pretty_func': self.pretty_time},  # time in ns up to 200 000 days.
            SIZE: {'type': ARRAY_UINT32, 'pretty_func': pretty_sz},  # sizes up to 4 Gbytes
        }
        self.STAT_DEFAULT = {
            name: array(params['type']) for name, params in self.DIMENSIONS.items()
        }

        self.start_ns = time_ns()

        self.time_units = time_units
        self.time_threshold_ns = time_threshold_ms * 10**6
        self.ok = success_statuses

        self.stat = {}  # stat[request type][status]
        # todo implement cache
        # To select the best approach cache compare benchmarks for two versions:
        #  1) fill cache in log() and use in filter()/reduce()
        #  2) fill cache in filter() and use in reduce()
        # self.stat_by_group = {}  # cache
        # self.stat_by_request = {}  # cache

    def group_name_by_status(self, status):
        """
        All this statuses should be in GROUPS
        """
        if status in self.ok:
            return SUCCESS_GROUP
        else:
            return FAIL_GROUP

    def log(self, status, elapsed, request_name, response_size):
        """
        Add result to the report

        :param status: HTTP response status
        :param elapsed: Request-Response time, ns
        :param request_name: Request name or None
        :param response_size: Response body size
        """
        self.stat.setdefault(request_name, {}).setdefault(status, deepcopy(self.STAT_DEFAULT))
        self.stat[request_name][status][TIME].append(elapsed // TIME_DENOMINATOR)
        self.stat[request_name][status][SIZE].append(response_size)

    @property
    def total_elapsed_ns(self):
        return time_ns() - self.start_ns

    def reduce(
            self,
            reduce_func,
            dimension_name: str,
            status_group_filter: str = None,
            request_name_filter: str = None
    ) -> int:
        """
        Reduce the dimension by group and/or request_name with the reduce_func
        Returns dict {'time':, 'size':}
        """
        result = 0
        for request_name in self.stat:
            if request_name == request_name_filter or request_name_filter is None:
                for status in self.stat[request_name]:
                    if self.group_name_by_status(status) == status_group_filter or status_group_filter is None:
                        result += reduce_func(self.stat[request_name][status][dimension_name])
        return result

    def filter(
            self,
            dimension_name: str,
            status_group_filter: str = None,
            request_name_filter: str = None
    ) -> array:
        """
        Filter the dimension by group and/or request_name,
        Returns array with values from dimention_name
        """
        dimension = array(self.DIMENSIONS[dimension_name]['type'])
        for request_name in self.stat:
            if request_name == request_name_filter or request_name_filter is None:
                for status in self.stat[request_name]:
                    if self.group_name_by_status(status) == status_group_filter or status_group_filter is None:
                        dimension += self.stat[request_name][status][dimension_name]
        return dimension

    @staticmethod
    def dimension_stat_report(dimension_values: array, pretty_func) -> str:
        if not dimension_values:
            return ''
        return ', '.join([
            f'Mean: {pretty_func(statistics.mean(dimension_values))}',
            f'min: {pretty_func(min(dimension_values))}',
            f'max: {pretty_func(max(dimension_values))}',
        ])

    def filtered_report(
            self,
            status_group_filter: str = None,
            request_name_filter: str = None
    ) -> str:
        """
        Filter by group and/or request_name.
        Returns report str with stats for all dimensions.
        """
        result = []
        for dimension_name in self.DIMENSIONS:
            dimension = self.filter(dimension_name, status_group_filter, request_name_filter)
            if dimension:
                result.append(
                    self.dimension_stat_report(
                        dimension,
                        self.DIMENSIONS[dimension_name]['pretty_func']
                    )
                )
        if not result:
            return 'No such requests'
        return '\n'.join(result)

    def statuses_report(self, request_name_filter: str = None) -> str:
        return ', '.join([f'{group} {self.reduce(len, TIME, group, request_name_filter)}'
                          for group in GROUPS])

    def pretty_time(self, elapsed: int, paint: bool = True):
        return self.pretty_ns(elapsed * TIME_DENOMINATOR, paint)

    def pretty_ns(self, elapsed_ns: int, paint: bool = True):
        result = pretty_ns(elapsed_ns, self.time_units)
        if elapsed_ns > self.time_threshold_ns and paint:
            return red(result)
        else:
            return result

    def report(self):
        size_sum = self.reduce(sum, SIZE)
        total_ns = self.reduce(sum, TIME)
        total_num = self.reduce(len, TIME)
        elapsed_sec = total_ns / (10 ** 9)
        infinity = chr(0x221E)
        total_line = ' '.join([
            f'Got `{total_num}` responses',
            f'in `{self.pretty_time(total_ns, paint=False)}`,',
            f'`{round(total_num / elapsed_sec) if elapsed_sec > 0 else infinity} op/sec`,',
            f'{pretty_sz(size_sum)},',
            f'{pretty_sz(size_sum // elapsed_sec) if elapsed_sec > 0 else infinity}/sec',
        ])
        by_group = []
        for status_group in GROUPS:
            by_group.append(f'''#### {status_group}: {self.reduce(len, TIME, status_group)} 
{self.filtered_report(status_group)}
''')
        by_group = '\n'.join(by_group)
        by_request = []
        for request_name in self.stat:
            by_request.append(f'''### {request_name}: {self.statuses_report(request_name)}
{self.filtered_report(None, request_name)}
''')
        by_request = '\n'.join(by_request)
        return f'''{total_line}
        
{by_group}

{by_request}'''



from tests.stdout_capture import CaptureOutput
import unittest
from bombard.main import campaign
from tests.fake_args import FakeArgs
import bombard


class TestVersion(unittest.TestCase):
    def testVersion(self):
        with CaptureOutput() as captured:
            campaign(FakeArgs(version=True))

        with open(f'bombard/LICENSE.txt') as license:
            self.assertIn(
                license.readline(),
                captured.output
                )
        self.assertIn(
            bombard.version(),
            captured.output
            )



import os.path
from bombard.args import EXAMPLES_PREFIX, CAMPAIGN_FILE_NAME, DIR_DESC_FILE_NAME
from bombard.terminal_colours import red
from bombard.show_descr import markdown_for_terminal
import bombard


def expand_relative_file_name(file_name):
    """
    Replace RELATIVE_PREFIX with package folder so bombard script can use internal examples without full path spec
    """
    if file_name.strip().startswith(EXAMPLES_PREFIX):
        # resource_string(__name__, args.file_name[1:])  # recommended use resource to be zipfile compatible. but this is a pain for !include
        return os.path.join(os.path.dirname(bombard.__file__), 'examples', file_name[len(EXAMPLES_PREFIX):])
    else:
        return file_name


def get_campaign_file_name(args):
    if args.example is not None:
        if args.file_name != CAMPAIGN_FILE_NAME:
            print(red(f'--example option found - ignoring campaign file name "{args.file_name}".'))
        args.file_name = EXAMPLES_PREFIX + args.example
        if not args.file_name.endswith('.yaml'):
            args.file_name += '.yaml'
    if args.examples:
        if args.file_name != CAMPAIGN_FILE_NAME:
            print(red(f'--examples option found - ignoring campaign file name "{args.file_name}".'))
        if args.example is not None:
            print(red('Please do not use --example and --examples options simultaneously.'))
        args.file_name = EXAMPLES_PREFIX

    return expand_relative_file_name(args.file_name)


def show_folder(folder_path):
    file_name = os.path.join(folder_path, DIR_DESC_FILE_NAME)
    if not os.path.isfile(file_name):
        print(f'\nNo {DIR_DESC_FILE_NAME} in folder {folder_path}. \nFolder content:\n')
        for name in os.listdir(folder_path):
            print(name)
    else:
        print(f'\n{folder_path}:\n')
        with open(file_name, 'r') as desc:
            print(markdown_for_terminal(desc.read()))


